
class ChapterBuses {

	constructor() {
	}

	static function ID();
	static function LoadChapter(main_instance);
}

/*static*/ function ChapterBuses::ID()
{
	return "buses";
}

/*static*/ function ChapterBuses::LoadChapter(main_instance)
{
	// Initialization code
	main_instance.AddStep(CodeStep( function(table) {

		// Grow this town:
		table.town <- 6; // Luham

		// Other towns to optionally connect
		table.town_other <- 0; // Wanpool

		// Store initial population
		table.initial_population <- GSTown.GetPopulation(table.town); 
	}));

	// 5.1 - Introduction
	main_instance.AddStep(MessageWindowStep(GSText(GSText.STR_BUSES_5_1_1_INTRO), WAIT, TableKey("town")));
	main_instance.AddStep(CodeStep( function(table) {
		GSViewport.ScrollTo(GSTown.GetLocation(table.town)); // scroll viewport to town
	}));
	main_instance.AddStep(MessageWindowStep(GSText(GSText.STR_BUSES_5_1_2_GROWTH_INFO), WAIT));
	main_instance.AddStep(MessageWindowStep(GSText(GSText.STR_BUSES_5_1_3_GROWTH_INFO), NO_WAIT, TableKey("town")));
	main_instance.AddStep(WaitOnWindowStep(GSWindow.WC_TOWN_VIEW, TableKey("town"), WAIT_ON_OPEN));
	main_instance.AddStep(MessageWindowStep(GSText(GSText.STR_BUSES_5_1_4_TOWN_VIEW), WAIT));

	// 5.2 - Build Bus Stations
	main_instance.AddStep(MessageWindowStep(GSText(GSText.STR_BUSES_5_2_1_BUS_STATION), NO_WAIT));
	main_instance.AddStep(GUIHighlightStep(GSWindow.WC_MAIN_TOOLBAR, 0, GSWindow.WID_TN_ROADS, NO_WAIT));
	main_instance.AddStep(WaitOnWindowStep(GSWindow.WC_BUILD_TOOLBAR, 1, WAIT_ON_OPEN));
	main_instance.AddStep(MessageWindowStep(GSText(GSText.STR_BUSES_5_2_2_BUS_STATION), NO_WAIT));
	main_instance.AddStep(GUIHighlightStep(GSWindow.WC_BUILD_TOOLBAR, 1, GSWindow.WID_ROT_BUS_STATION, NO_WAIT));
	main_instance.AddStep(WaitOnWindowStep(GSWindow.WC_BUS_STATION, 1, WAIT_ON_OPEN));
	main_instance.AddStep(MessageWindowStep(GSText(GSText.STR_BUSES_5_2_3_BUILD), NO_WAIT, TableKey("town")));
	main_instance.AddStep(CodeStep( function(table) {
		// <wait until the user builds 4 road stops and all are connected
		Common.WaitFor(ChapterBuses.GetBusStops, [table.town], GSText(GSText.STR_BUSES_NOTICE_WAITING_FOR_BUS_STOPS, table.town));
		// bus_stops = squirrel array with >= 4 bus stops
		table.bus_stops <- ChapterBuses.GetBusStops(table.town);
		// make two pairs of bus stops that provide longest total distance
		// station_pair1 = array
		local tmp = ChapterBuses.GetOptimalStopPairs(table.bus_stops);
		if (tmp == null) Log.Error("Failed to pick best station pairs");
		table.station_pair1_stop1 <- tmp[0][0];
		table.station_pair1_stop2 <- tmp[0][1];
		table.station_pair2_stop1 <- tmp[1][0];
		table.station_pair2_stop2 <- tmp[1][1];
	}));

	// 5.3 - Build depot
	main_instance.AddStep(MessageWindowStep(GSText(GSText.STR_BUSES_5_3_1_DEPOT), NO_WAIT));
	main_instance.AddStep(CodeStep( function(table) {
		// <wait until the user builds 4 road stops and all are connected
		Common.WaitFor(ChapterBuses.GetConnectedDepot, [table.town, table.bus_stops[0]], GSText(GSText.STR_BUSES_NOTICE_WAITING_FOR_DEPOT, table.town));
		table.depot <- ChapterBuses.GetConnectedDepot(table.town, table.bus_stops[0]);
	}));

	// 5.4 - Bus + Orders I
	main_instance.AddStep(CodeStep( function(table) {
	}));
	main_instance.AddStep(MessageWindowStep(GSText(GSText.STR_BUSES_5_4_1_ORDERS), NO_WAIT, TableKey("station_pair1_stop1"), TableKey("station_pair1_stop2")));
	main_instance.AddStep(CodeStep( function(table) {
		// wait for bus to be built
		Common.WaitFor(Common.GetVehicle, [GSVehicle.VT_ROAD, Helper.GetPAXCargo()], GSText(GSText.STR_BUSES_NOTICE_WAITING_FOR_BUS, table.town));
		table.pair1_bus <- Common.GetVehicle(GSVehicle.VT_ROAD, Helper.GetPAXCargo());
		// wait for bus to have orders
		Common.WaitForOrderToStations(table.pair1_bus, [table.station_pair1_stop1, table.station_pair1_stop2], GSText(GSText.STR_BUSES_NOTICE_WAITING_FOR_BUS_ORDERS, table.pair1_bus, table.station_pair1_stop1, table.station_pair1_stop2));
	}));
	main_instance.AddStep(MessageWindowStep(GSText(GSText.STR_BUSES_5_4_2_START), NO_WAIT));
	main_instance.AddStep(CodeStep( function(table) {
		Common.WaitFor(Common.VehicleStarted_WaitCondition, [table.pair1_bus], GSText(GSText.STR_BUSES_NOTICE_WAITING_FOR_START_BUS, table.pair1_bus));
	}));

	// 5.5 - Bus + Orders II
	main_instance.AddStep(MessageWindowStep(GSText(GSText.STR_BUSES_5_5_1_ORDERS), NO_WAIT, TableKey("station_pair2_stop1"), TableKey("station_pair2_stop2")));
	main_instance.AddStep(CodeStep( function(table) {
		// wait for bus #2 to be built
		Common.WaitFor(Common.GetVehicle, [GSVehicle.VT_ROAD, Helper.GetPAXCargo(), [table.pair1_bus]], GSText(GSText.STR_BUSES_NOTICE_WAITING_FOR_BUS2, table.town));
		table.pair2_bus <- Common.GetVehicle(GSVehicle.VT_ROAD, Helper.GetPAXCargo(), [table.pair1_bus]);
		// wait for bus to have orders
		Common.WaitForOrderToStations(table.pair2_bus, [table.station_pair2_stop1, table.station_pair2_stop2], GSText(GSText.STR_BUSES_NOTICE_WAITING_FOR_BUS_ORDERS, table.pair2_bus, table.station_pair2_stop1, table.station_pair2_stop2));
	}));
	main_instance.AddStep(MessageWindowStep(GSText(GSText.STR_BUSES_5_5_2_START), NO_WAIT));
	main_instance.AddStep(CodeStep( function(table) {
		Common.WaitFor(Common.VehicleStarted_WaitCondition, [table.pair2_bus], GSText(GSText.STR_BUSES_NOTICE_WAITING_FOR_START_BUS, table.pair2_bus));
	}));

	// 5.6 - Clone
	main_instance.AddStep(MessageWindowStep(GSText(GSText.STR_BUSES_5_6_1_CLONE), WAIT));

	// 5.6
	main_instance.AddStep(CodeStep( function(table) {
		// target: grow 100 additional inhabitants from now or at least minimum to 50 inhabitants more than initial.
		local pop = max(GSTown.GetPopulation(table.town) + 100, table.initial_population + 50);
		table.target_population <- (pop / 50) * 50 + (pop % 50 != 0? 50 : 0); // round up to closest number that is a factor of 50
	}));
	main_instance.AddStep(MessageWindowStep(GSText(GSText.STR_BUSES_5_7_1_CONCLUSION), NO_WAIT, TableKey("town"), TableKey("target_population")));
	main_instance.AddStep(CodeStep( function(table) {
		Common.WaitFor(ChapterBuses.PopulationWaitCondition, [table.town, table.target_population], GSText(GSText.STR_BUSES_NOTICE_WAITING_FOR_POPULATION, table.town, table.target_population));
		table.reached_population <- max(GSTown.GetPopulation(table.town), table.target_population); // show target number even if pop decreased during the tick when the message is shown, but credit the player for any extra inhabitant.
	}));
	main_instance.AddStep(MessageWindowStep(GSText(GSText.STR_BUSES_5_7_2_CONCLUSION), WAIT, TableKey("initial_population"), TableKey("reached_population")));
	main_instance.AddStep(MessageWindowStep(GSText(GSText.STR_BUSES_5_7_3_CONCLUSION), WAIT, TableKey("town_other"), TableKey("town")));
}

/* static */ function ChapterBuses::StationDistanceToTown(station_id, town_id)
{
	return GSMap.DistanceManhattan(GSStation.GetLocation(station_id), GSTown.GetLocation(town_id));
}

/* static */ function ChapterBuses::GetBusStationTile(station_id)
{
	local st_tiles = GSTileList_StationType(station_id, GSStation.STATION_BUS_STOP);
	return st_tiles.Begin();
}

/* static */ function ChapterBuses::GetBusStops(town)
{
	local cm = GSCompanyMode(HUMAN_COMPANY);
	GSRoad.SetCurrentRoadType(GSRoad.ROADTYPE_ROAD);

	local st_list = GSStationList(GSStation.STATION_BUS_STOP);

	// Filter out only Human player stations
	st_list.Valuate(GSStation.GetOwner);
	st_list.KeepValue(HUMAN_COMPANY);
	
	// Filter stations to only contain stations near town
	st_list.Valuate(ChapterBuses.StationDistanceToTown, town);
	st_list.KeepBelowValue(20);

	// Require at least 4 stations
	if(st_list.Count() < 4) return -1;

	// Ensure that all stations are inter-connected
	local one_station = st_list.Begin();
	local one_station_tile = ChapterBuses.GetBusStationTile(one_station);

	local all_stations_connected = true;
	local i = -1;
	foreach(st, _ in st_list)
	{
		i++;
		if(st == one_station) continue;

		local st_tile = ChapterBuses.GetBusStationTile(st);
		if(!Road.AreDistantRoadTilesConnected(
				one_station_tile, 
				st_tile))
		{
			all_stations_connected = false;
			Log.Info("from " + GSStation.GetName(one_station) + " to " + GSStation.GetName(st) + ": not connected", Log.LVL_DEBUG);
			break;
		}

		Log.Info("from " + GSStation.GetName(one_station) + " to " + GSStation.GetName(st) + ": connected", Log.LVL_DEBUG);
	}

	if (!all_stations_connected) return -1;

	// Convert st_list into a squirrel array so that it can be saved in save games
	local st_array = [];
	foreach(st, _ in st_list)
	{
		st_array.append(st);
	}

	return st_array;
}

/* static */ function ChapterBuses::GetOptimalStopPairs(stops_array)
{
	// We assume there is 4 bus stops. If there is more then some will get ignored.

	local solutions = [];
	local best_solution = -1;
	local best_score = 0;
	for(local i = 1; i < 4; i++) // i give the pair to combine with index 0 to form the first pair
	{

		local solution = [
				[stops_array[0], stops_array[i]],
				[stops_array[i == 1? 2 : 1], stops_array[i == 2? 3 : 2]]
		];

		local dist1 = GSMap.DistanceManhattan(ChapterBuses.GetBusStationTile(solution[0][0]), ChapterBuses.GetBusStationTile(solution[0][1]));
		local dist2 = GSMap.DistanceManhattan(ChapterBuses.GetBusStationTile(solution[1][0]), ChapterBuses.GetBusStationTile(solution[1][1]));

		local score = dist1 + dist2 - abs(dist1 - dist2); // last term favor solutions with two half-long legs rather than one long and one short.

		Log.Info("Score: " + score, Log.LVL_DEBUG);

		if (score > best_score)
		{
			best_solution = i - 1; // use array index
			best_score = score;
			Log.Info("New best solution: " + best_solution + " at score " + score, Log.LVL_DEBUG);
		}

		solutions.append(solution);
	}

	if (best_solution == -1) return null;

	Log.Info("Return best solution", Log.LVL_DEBUG);
	return solutions[best_solution];
}

/* static */ function ChapterBuses::GetConnectedDepot(town, station_id)
{
	local cm = GSCompanyMode(HUMAN_COMPANY);
	GSRoad.SetCurrentRoadType(GSRoad.ROADTYPE_ROAD);

	// Get tiles to search
	local tiles = Tile.MakeTileRectAroundTile(GSTown.GetLocation(town), 20);

	// Filter tiles for a depot owned by human company
	tiles.Valuate(GSRoad.IsRoadDepotTile);
	tiles.KeepValue(1);

	tiles.Valuate(GSTile.GetOwner);
	tiles.KeepValue(HUMAN_COMPANY);

	// No depot?
	if(tiles.IsEmpty()) 
	{
		Log.Info("No depot found", Log.LVL_DEBUG);
		return -1;
	}

	// Get station tile
	local st_tiles = GSTileList_StationType(station_id, GSStation.STATION_BUS_STOP);
	local st_tile = st_tiles.Begin();

	// Pick a depot
	foreach(depot_tile, _ in tiles)
	{
		if(Road.AreDistantRoadTilesConnected(depot_tile, st_tile))
		{
			Log.Info("Connected depot found", Log.LVL_DEBUG);
			// There is a road from the depot to the station
			return depot_tile;
		}

		Log.Info("Depot at " + Tile.GetTileString(depot_tile) + " is not connected", Log.LVL_DEBUG);
	}

	Log.Info("No connected depot found", Log.LVL_DEBUG);
	// No connected depot found
	return -1;
}

/* static */ function ChapterBuses::PopulationWaitCondition(town, target)
{
	return GSTown.GetPopulation(town) < target? -1: 1;
}
